package com.aliyun.odps.spark.examples.makedata_recsys
import com.aliyun.odps.TableSchema
import com.aliyun.odps.data.Record
import com.aliyun.odps.spark.examples.myUtils.{ParamUtils, env}
import org.apache.spark.sql.{DataFrame, Row, SparkSession}
import org.apache.spark.sql.functions._
import org.apache.spark.sql.types._
import org.apache.spark.storage.StorageLevel
import com.jayway.jsonpath.JsonPath

import scala.util.Random
import scala.collection.mutable.ArrayBuffer
import org.apache.log4j.{Level, Logger}

object video_dssm_sampler {
  private val logger = Logger.getLogger(this.getClass)

  // 配置参数
  private val CONFIG = Map(
    "shuffle.partitions" -> "200",
    "memory.fraction" -> "0.8",
    "default.parallelism" -> "200"
  )

  case class ProcessingStats(
                              startTime: Long = System.currentTimeMillis(),
                              var positiveSamplesCount: Long = 0,
                              var negativeSamplesCount: Long = 0
                            ) {
    def logStats(): Unit = {
      val duration = (System.currentTimeMillis() - startTime) / 1000
      logger.info(s"""
                     |Processing Statistics:
                     |Duration: ${duration}s
                     |Positive Samples: $positiveSamplesCount
                     |Negative Samples: $negativeSamplesCount
                     |Total Samples: ${positiveSamplesCount + negativeSamplesCount}
        """.stripMargin)
    }
  }

  def createSparkSession(appName: String): SparkSession = {
    val spark = SparkSession.builder()
      .appName(appName)
      .enableHiveSupport()
      .config("spark.sql.shuffle.partitions", CONFIG("shuffle.partitions"))
      .config("spark.memory.fraction", CONFIG("memory.fraction"))
      .config("spark.default.parallelism", CONFIG("default.parallelism"))
      .getOrCreate()

    // 设置日志级别
    spark.sparkContext.setLogLevel("WARN")
    spark
  }

  def funcPositive(record: Record, schema: TableSchema): Row = {
    //println(s"schema.getColumns.toString = ${schema.getColumns.toString}")
    //println(s"Record: ${record.toString}")
    Row(
      //record.getString("dt"),
      //record.getString("hh"),
      record.getString("vid_left"),
      record.getString("vid_right"),
      record.getString("extend"),
      record.getString("view_24h"),
      record.getString("total_return_uv"),
      record.getString("ts_right"),
      record.getString("apptype"),
      record.getString("pagesource"),
      record.getString("mid"))
  }
  def funcVids(record: Record, schema: TableSchema): String = {
    record.getString("vid")
  }
  def funcTagFeatures(record: Record, schema: TableSchema): (String, String) = {
    (record.getString("vid"), record.getString("feature"))
  }
  def funcStatFeatures(record: Record, schema: TableSchema): (String, String) = {
    (record.getString("vid"), record.getString("feature"))
  }
  // 获取类别特征
  def funcCategoryFeatures(record: Record, schema: TableSchema): (String, String, String, String) = {
    val feature = record.getString("feature")
    val category1 = JsonPath.read[String](feature, "$.category1")
    val category2 = JsonPath.read[String](feature, "$.category2_1")
    (record.getString("vid"), category1, category2, feature)
  }
  // 1. 首先定义L1和L2类别统计特征的读取函数
  def funcCatStatFeatures(record: Record, schema: TableSchema): (String, String) = {
    (record.getString("category1"), record.getString("feature"))  // 或 category2 取决于表
  }



  def generateNegativeSamples(spark: SparkSession, dt: String, outputPath: String): Unit = {
    val stats = ProcessingStats()

    try {
      logger.info(s"Starting negative sample generation for date: $dt")
      val sc = spark.sparkContext

      // 2 读取odps+表信息
      val odpsOps = env.getODPS(sc)
      // 1. 获取正样本对数据

      val schema = StructType(Array(
        //StructField("dt", StringType, true),
        //StructField("hh", StringType, true),
        StructField("vid_left", StringType, true),
        StructField("vid_right", StringType, true),
        StructField("extend", StringType, true),
        StructField("view_24h", StringType, true),
        StructField("total_return_uv", StringType, true),
        StructField("ts_right", StringType, true),
        StructField("apptype", StringType, true),
        StructField("pagesource", StringType, true),
        StructField("mid", StringType, true)
      ))


      // logger.info(s"start read positivePairs for date $dt")
      val partition = s"dt=20241124"
      val rdd = odpsOps.readTable(
        project = "loghubods",
        table = "alg_dssm_sample",
        partition = s"dt=20241124,hh=08",
        transfer = funcPositive,
        numPartition = CONFIG("shuffle.partitions").toInt
      ).sample(false, 0.001) // 随机抽样千分之一的数据
       .persist(StorageLevel.MEMORY_AND_DISK)
      println("开始执行partiton:" + partition)

      val positivePairs = spark.createDataFrame(rdd, schema).persist(StorageLevel.MEMORY_AND_DISK)
      stats.positiveSamplesCount = positivePairs.count()
      logger.info(s"start read vid list for date $dt")

      // 2. 获取所有可用的vid列表
      val allVids = time({
        odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_tag_feature",
          partition = s"dt='$dt'",
          transfer = funcVids,
          numPartition = CONFIG("shuffle.partitions").toInt
        ).collect().toSet
      }, "Loading all vids")


      val allVidsBroadcast = spark.sparkContext.broadcast(allVids)
      // 注册UDF
      spark.udf.register("generateNegativeVids", generateNegativeVidsUDF)
      // 3. 定义UDF函数来生成负样本
      def generateNegativeVidsUDF = udf((vid_left: String, vid_right: String) => {
        val localAllVids = allVidsBroadcast.value
        val negativeVids = Random.shuffle(localAllVids - vid_left - vid_right).take(2) //20
        negativeVids.toArray
      })
      // 4. 生成负样本
      val negativeSamplesDF = time({
        val df = positivePairs
          .select("vid_left", "vid_right")
          .withColumn("negative_vids", generateNegativeVidsUDF(col("vid_left"), col("vid_right")))
          .select(
            col("vid_left"),
            explode(col("negative_vids")).as("vid_right"),
            lit(null).as("extend"),
            lit(null).as("view_24h"),
            lit(null).as("total_return_uv"),
            lit(null).as("ts_right"),
            lit(null).as("apptype"),
            lit(null).as("pagesource"),
            lit(null).as("mid")
          )

        stats.negativeSamplesCount = df.count()
        df
      }, "Generating negative samples")

      // 5. 合并正负样本
      val allSamples = positivePairs
        .withColumn("label", lit(1))
        .withColumn("logid", concat(lit("pos_"), monotonically_increasing_id()))
        .union(
          negativeSamplesDF
            .withColumn("label", lit(0))
            .withColumn("logid", concat(lit("neg_"), monotonically_increasing_id()))
        ).persist(StorageLevel.MEMORY_AND_DISK)

      // 6. 获取左侧特征
      // 读取L1类别统计特征
      val l1CatStatFeatures = {
        val rdd = odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_l1_cat_stat_feature",
          partition = s"dt='$dt'",
          transfer = funcCatStatFeatures,
          numPartition = CONFIG("shuffle.partitions").toInt
        )
        val schema = StructType(Array(
          StructField("category1", StringType, true),
          StructField("cate_l1_feature", StringType, true)
        ))
        spark.createDataFrame(rdd.map(t => Row(t._1, t._2)), schema)
      }

      // 读取L2类别统计特征
      val l2CatStatFeatures = {
        val rdd = odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_l2_cat_stat_feature",
          partition = s"dt='$dt'",
          transfer = funcCatStatFeatures,
          numPartition = CONFIG("shuffle.partitions").toInt
        )
        val schema = StructType(Array(
          StructField("category2", StringType, true),
          StructField("cate_l2_feature", StringType, true)
        ))
        spark.createDataFrame(rdd.map(t => Row(t._1, t._2)), schema)
      }

      val categoryData = {
        // 从 t_vid_tag_feature 表读取数据
        val rdd = odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_tag_feature",
          partition = s"dt='$dt'",
          transfer = funcCategoryFeatures,  // 使用之前定义的 funcCategoryFeatures 函数
          numPartition = CONFIG("shuffle.partitions").toInt
        )

        // 定义 schema
        val schema = StructType(Array(
          StructField("vid", StringType, true),
          StructField("category1", StringType, true),
          StructField("category2", StringType, true),
          StructField("feature", StringType, true)
        ))

        // 将 RDD 转换为 DataFrame
        spark.createDataFrame(rdd.map(t => Row(t._1, t._2, t._3, t._4)), schema)
      }


      // 获取tag特征

      val tagFeatures = {
        val rdd = odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_tag_feature",
          partition = s"dt='$dt'",
          transfer = funcTagFeatures,
          numPartition = CONFIG("shuffle.partitions").toInt
        )
        val schema = StructType(Array(
          StructField("vid", StringType, true),
          StructField("vid_left_tag_feature", StringType, true)
        ))
        spark.createDataFrame(rdd.map(t => Row(t._1, t._2)), schema)
      }


      // 获取统计特征
      val statFeatures = {
        val rdd = odpsOps.readTable(
          project = "loghubods",
          table = "t_vid_stat_feature",
          partition = s"dt='$dt'",
          transfer = funcStatFeatures,
          numPartition = CONFIG("shuffle.partitions").toInt
        )
        val schema = StructType(Array(
          StructField("vid", StringType, true),
          StructField("vid_left_stat_feature", StringType, true)
        ))
        spark.createDataFrame(rdd.map(t => Row(t._1, t._2)), schema)
      }


      val categoryWithStats = categoryData
        .join(broadcast(l1CatStatFeatures), categoryData("category1") === l1CatStatFeatures("category1"), "left")
        .join(broadcast(l2CatStatFeatures), categoryData("category2") === l2CatStatFeatures("category2"), "left")
        .select(
          col("vid"),
          col("cate_l1_feature").as("vid_left_cate_l1_feature"),
          col("cate_l2_feature").as("vid_left_cate_l2_feature")
        )

      val vidLeftFeatures = allSamples
        .join(broadcast(tagFeatures), col("vid_left") === tagFeatures("vid"), "left")
        .drop(tagFeatures("vid"))
        .join(broadcast(statFeatures), col("vid_left") === statFeatures("vid"), "left")
        .drop(statFeatures("vid"))
        .join(broadcast(categoryWithStats), col("vid_left") === categoryWithStats("vid"), "left")
        .drop(categoryWithStats("vid"))
        .select(
          allSamples("*"),
          col("vid_left_tag_feature"),
          col("vid_left_stat_feature"),
          col("vid_left_cate_l1_feature"),
          col("vid_left_cate_l2_feature")
        )
        .persist(StorageLevel.MEMORY_AND_DISK)

      // 合并所有右侧特征并生成最终结果
      val result = vidLeftFeatures
      //  .join(broadcast(tagFeatures), col("vid_right") === tagFeatures("vid"), "left")
      //  .join(broadcast(statFeatures), col("vid_right") === statFeatures("vid"), "left")
      //  .join(broadcast(categoryData), col("vid_right") === categoryData("vid"), "left")

      // 保存结果到HDFS
      val resultWithDt = result.withColumn("dt", lit(s"$dt"))
      resultWithDt.write
        .mode("overwrite")
        .partitionBy("dt")
        .parquet(s"$outputPath/dt=$dt")

      logger.info(s"Results saved to: $outputPath/dt=$dt")

      // 8. 清理缓存
      positivePairs.unpersist()
      allSamples.unpersist()
      vidLeftFeatures.unpersist()


      // 输出统计信息
      stats.logStats()

    } catch {
      case e: Exception =>
        logger.error(s"Error processing data for date $dt: ${e.getMessage}")
        e.printStackTrace()
        throw e
    }
  }

  private def time[R](block: => R, name: String): R = {
    val t0 = System.nanoTime()
    val result = block
    val t1 = System.nanoTime()
    logger.info(s"$name took: ${(t1 - t0) / 1e9d} seconds")
    result
  }

  def main(args: Array[String]): Unit = {
    if (args.length < 2) {
      println("Usage: NegativeSampleGenerator <date> <output_path>")
      System.exit(1)
    }

    val dt = args(0)
    val outputPath = args(1)

    val spark = createSparkSession("NegativeSampleGenerator")

    try {
      generateNegativeSamples(spark, dt, outputPath)
    } finally {
      spark.stop()
    }
  }
}

// 运行
// 上传代码
// ssh 192.168.141.208
// git clone
// mvn clean install
// /opt/apps/SPARK2/spark-2.4.8-hadoop3.2-1.0.8/bin/spark-class2 org.apache.spark.deploy.SparkSubmit   --class com.aliyun.odps.spark.examples.makedata_recsys.video_dssm_sampler   --master yarn --driver-memory 1G --executor-memory 2G --executor-cores 1 --num-executors 16   ./target/spark-examples-1.0.0-SNAPSHOT-shaded.jar    20241120 /dw/recommend/dssm_model/